<?php
/*
  This code is part of FusionDirectory (http://www.fusiondirectory.org/)
  Copyright (C) 2007  Fabian Hicker
  Copyright (C) 2011-2018  FusionDirectory

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
*/

class MailQuotaAttribute extends UnitIntAttribute
{
  /**
   * @var int Quota usage in MiB, -1 if unknown
   */
  private $quotaUsage;
  /**
   * @var string Quota limit in MiB, empty string if unlimited
   *
   * The same as $this->value, unless there is an inherited group limit
   */
  private $quotaLimit;

  function __construct ($label, $description, $ldapName, $required, $min = FALSE, $max = FALSE, $defaultValue = "", $acl = "")
  {
    $units = [
      1         => _('MiB'),
      1024 ** 1 => _('GiB'),
      1024 ** 2 => _('TiB'),
    ];
    parent::__construct($label, $description, $ldapName, $required, $units, $min, $max, $defaultValue, $acl);
  }

  function setMailMethod ($mailMethod)
  {
    try {
      /* Read quota */
      $this->setValue((string) $mailMethod->getQuota($this->getValue()));
      $this->quotaUsage = $mailMethod->getQuotaUsage();
      $this->quotaLimit = $mailMethod->getQuotaLimit($this->getValue());
    } catch (MailMethodError $e) {
      $error = new SimplePluginError(
        $this,
        htmlescape(sprintf(
          _('Cannot read quota settings: %s'),
          $e->getMessage()
        )),
        0,
        $e
      );
      $error->display();
    }
  }

  function renderAttribute (array &$attributes, bool $readOnly, bool $readable, bool $writable)
  {
    parent::renderAttribute($attributes, $readOnly, $readable, $writable);
    $attributes[$this->getLdapName().'_usage'] = [
      'htmlid'        => $this->getHtmlId().'_usage',
      'label'         => '{literal}'._('Quota usage').'{/literal}',
      'description'   => _('Part of the quota which is used'),
      'input'         => static::quotaToImage($this->quotaUsage, $this->quotaLimit),
      'subattribute'  => FALSE,
      'required'      => FALSE,
      'readable'      => $readable,
      'writable'      => $writable,
    ];
  }

  static function quotaToImage (int $use, string $quota): string
  {
    if ($use == -1) {
      return '&nbsp;&nbsp;'.htmlescape(_('Unknown'));
    } elseif (empty($quota)) {
      return '&nbsp;&nbsp;'.htmlescape(sprintf(_('Unlimited (%dMiB used)'), $use));
    } else {
      $usage = round(($use / $quota) * 100);
      $alt = htmlescape(sprintf(_('%d%% (%dMiB)'), $usage, $use));
      return '<img src="progress.php?x=100&amp;y=17&amp;p='.$usage.'" alt="'.$alt.'" title="'.$alt.'"/>';
    }
  }
}
