<?php
/*
  This code is part of FusionDirectory (http://www.fusiondirectory.org/)
  Copyright (C) 2003-2010  Cajus Pollmeier
  Copyright (C) 2011-2021  FusionDirectory

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
*/

/*!
 * \brief Error class which can be thrown by mail method when error occurs
 */
class MailMethodError extends FusionDirectoryError
{
}

class mailMethod
{
  /* Allow modification of account_ids for existing mail accounts */
  protected $modifyableMail   = TRUE;

  /* Allow modification of the mail server attribute existing mail accounts */
  protected $modifyableServer = TRUE;

  /* Create accounts with '/' instead of '.' */
  protected $useSlashesInId = FALSE;

  /* Features */
  protected $enableQuota            = TRUE;
  protected $enableVacation         = TRUE;
  protected $enableVacationRange    = FALSE;
  protected $enableGroupForwarding  = TRUE;
  protected $enableLocalOnly        = TRUE;
  protected $enableForwardOnly      = TRUE;
  protected $enableGroupACLs        = FALSE;

  /* Default values */
  /**
  * @var string
  */
  protected $quotaValue   = '';
  /**
  * @var int
  */
  protected $quotaUsage   = -1;

  /* Method internal */
  protected $type               = 'user';
  protected $account_id         = '';
  protected $initial_account_id = '';
  protected $connected          = FALSE;
  protected $parent             = NULL;
  protected $mailServer         = NULL;

  protected $acl_map = [
    ''            => 'none',
    'p'           => 'external post',
    'lrs'         => 'read',
    'lrsp'        => 'post',
    'lrspxte'     => 'delete',
    'lrsip'       => 'append',
    'lrswipkxte'  => 'write',
    'lrswipkxtea' => 'admin',
  ];

  protected $acl_mapping = [];

  /*! \brief  Constructs the mail class
      @param  Object  Plugin  The initator
      @param  String          Open "user" or "group" account.
   */
  function __construct ($parent, string $type = 'user')
  {
    $this->parent = $parent;

    if (!isset($this->parent->gosaMailServer)) {
      throw new FusionDirectoryException('mailMethod with invalid parent object initialized.');
    }

    $this->mailServer = $this->parent->gosaMailServer;

    if (!in_array($this->type, ['user','group'])) {
      trigger_error('Unknown mail class type used "'.$type.'".');
    } else {
      $this->type = $type;
    }

    $this->init();
  }

  /*! \brief  Intialize attributes and config settings.
   */
  protected function init ()
  {
    global $config;

    /* Create ACL map */
    foreach ($this->acl_map as $acl => $name) {
      $this->acl_mapping[$acl] = sprintf(_('%s (%s)'), _($name), $acl);
    }

    $this->build_account_id();
    $this->initial_account_id = $this->account_id;
  }

  /*! \brief  Connect to mail server or throw error
   */
  protected function connect ()
  {
    @DEBUG(DEBUG_MAIL, __LINE__, __FUNCTION__, __FILE__, get_class($this), '<b>MAIL: Connect method</b>');
    @DEBUG(DEBUG_MAIL, __LINE__, __FUNCTION__, __FILE__, $this->mailServer, '<b>MAIL: Current server</b>');

    $this->connected = TRUE;
  }

  public function isConnected (): bool
  {
    return $this->connected;
  }

  /*! \brief  Disconnect this method. Close services like imap connection.
              Not necessary for the base class.
   */
  public function disconnect ()
  {
    if ($this->isConnected()) {
      @DEBUG(DEBUG_MAIL, __LINE__, __FUNCTION__, __FILE__, get_class($this), '<b>MAIL: Disconnect method</b>');
      $this->connected = FALSE;
    }
  }

  protected function connectIfNeeded ()
  {
    if (!$this->isConnected()) {
      $this->connect();
    }
  }

  /*! \brief Checks data consistency
  */
  public function check (): array
  {
    return [];
  }

  /*! \brief  Returns true the current object represents a valid account
              (Some methods may check imap accounts here.)
      @return Boolean TRUE if this is a valid account else FALSE
  */
  public function accountExists (): bool
  {
    return TRUE;
  }

  public function isModifyableMail (): bool
  {
    return $this->modifyableMail;
  }

  public function isModifyableServer (): bool
  {
    return $this->modifyableServer;
  }

  public function getAccountId (): string
  {
    $this->build_account_id();
    return $this->account_id;
  }

  /*! \brief  Create a new account id
   */
  protected function build_account_id ()
  {
    $mail = '';
    if (isset($this->parent->mail)) {
      $mail = $this->parent->mail;
    }

    /* Create account_id */
    $acc_id = trim($mail);

    if ($this->account_id != $acc_id) {
      $this->account_id = $acc_id;
      @DEBUG(DEBUG_MAIL, __LINE__, __FUNCTION__, __FILE__, $acc_id, '<b>MAIL:</b> AccountID generated');
    }
  }

  /*! \brief  Creates a valid folder id for a given folder name.
               e.g. $folder_id = "INBOX/test"  &&  $this->account_id = "share/mailbox@fusiondirectory.org"
               will result in "share/mailbox/test@fusiondirectory.org"
              This function is mainly used to read and write folder permissions.
      @return String A valid folder id
   */
  protected function create_folder_id ($folder)
  {
    $delim = ($this->useSlashesInId ? '/' : '.');
    if (!empty($folder)) {
      $folder = trim(preg_replace("/^INBOX[\.\/]*/i", '', $folder));
    }
    if (!empty($folder)) {
      $folder = $delim.$folder;
    }

    /* We need to move the domain to the end
     * Example: user/john/Drafts@domain.com instead of user/john@domain.com/Drafts
    */
    $account  = $this->account_id;
    $suffix   = '';
    if (preg_match('/^(.+)(\@[^@]+)$/', $this->account_id, $m)) {
      $account  = $m[1];
      $suffix   = $m[2];
    }

    return trim($account.$folder.$suffix);
  }

  /*! \brief Saves sieve settings
   * @throws MailMethodError
   */
  public function saveSieveSettings ()
  {
  }

  /*! \brief  Creates or Updates the mailAccount represented by this class.
   * @throws MailMethodError
   */
  public function updateMailbox ()
  {
    @DEBUG(DEBUG_MAIL, __LINE__, __FUNCTION__, __FILE__, $this->account_id,
        "<b>MAIL: Update account</b> from server :".$this->mailServer);
  }

  /*! \brief  Removes the mailbox represented by this class,
               and update shared folder ACLs .
   * @throws MailMethodError
   */
  public function deleteMailbox ()
  {
    @DEBUG(DEBUG_MAIL, __LINE__, __FUNCTION__, __FILE__, $this->account_id,
        "<b>MAIL: Remove account</b> from server :".$this->mailServer);
  }

  /*! \brief  Returns whether the quota settings are enabled or not
      @return Boolean TRUE if enabled else FALSE
   */
  public function quotaEnabled (): bool
  {
    return $this->enableQuota;
  }

  /*! \brief  Returns the used quota
      @return Integer Quota used.
   */
  public function getQuotaUsage (): int
  {
    return -1;
  }

  /*! \brief  Returns the quota restrictions specific to this user.
      @return string quota value, numeric string or empty string
   */
  public function getQuota (string $quotaValue): string
  {
    return $quotaValue;
  }

  /*! \brief  Returns the quota restrictions specific or inherited by this user.
      @return string quota value, numeric string or empty string
   */
  public function getQuotaLimit (string $quotaValue): string
  {
    return $quotaValue;
  }

  /*! \brief  Sets the mail quota
   */
  public function setQuota (string $number)
  {
    if (!is_numeric($number)) {
      $number = (int) $number;
      if (!$number) {
        $number = 0;
      }
    }
    $this->quotaValue = $number;
  }

  /*! \brief  Returns the list of configured mailbox folders
      @return Array The mailbox folders.
   */
  public function getMailboxList (): array
  {
    return ['INBOX'];
  }

  /*! \brief  Returns whether the vacation is enabled
      @return Boolean TRUE, FALSE
   */
  public function vacationEnabled (): bool
  {
    return $this->enableVacation;
  }

  /*! \brief  Returns whether the vacation range is selectable or not
   */
  public function vacationRangeEnabled (): bool
  {
    return ($this->vacationEnabled() && $this->enableVacationRange);
  }

  /*! \brief  Returns whether forwarding to group non-members is possible
   */
  public function groupForwardingEnabled (): bool
  {
    return $this->enableGroupForwarding;
  }

  /*! \brief  Returns whether IMAP ACLs are supported for group members
   */
  public function groupAclsEnabled (): bool
  {
    return $this->enableGroupACLs;
  }

  /*! \brief  Returns whether forcing to only send/recieve local mail is possible
   */
  public function localOnlyEnabled (): bool
  {
    return $this->enableLocalOnly;
  }

  /*! \brief  Returns whether forwarding without storing the mail is possible
   */
  public function forwardOnlyEnabled (): bool
  {
    return $this->enableForwardOnly;
  }

  /*! \brief  Returns configuration of selected mail server
   * @throws MailMethodError
   * @return Array configuration
  */
  protected function getServerConfig (): array
  {
    $serverList = static::getMailServers();

    if (!count($serverList)) {
      throw new MailMethodError(htmlescape(_('There are no IMAP compatible mail servers defined!')));
    } elseif (!isset($serverList[$this->mailServer])) {
      throw new MailMethodError(htmlescape(_('Mail server for this account is invalid!')));
    } else {
      return $serverList[$this->mailServer];
    }
  }

  /*! \brief  Returns configured acls
   */
  public function getFolderACLs (): array
  {
    return [];
  }

  /*! \brief  Write ACLs back to imap or what ever
   */
  public function setFolderACLs (array $acls): bool
  {
    return TRUE;
  }

  /*! \brief  Returns a list of all possible acls.
      @return Array   ACLs.
  */
  public function getAclTypes (): array
  {
    return $this->acl_mapping;
  }

  public function additionalAttributes (string $acl): array
  {
    return [];
  }

  /*! \brief  Returns all mail servers configured in FusionDirectory
      @return Array  All useable mail servers
  */
  static public function getMailServers (): array
  {
    if (session::global_is_set('mailServers')) {
      return session::global_get('mailServers');
    }
    $methods = static::getMethodList();
    $servers = [];
    foreach ($methods as $class) {
      $servers = array_merge($servers, $class::getServerList());
    }
    session::global_set('mailServers', $servers);
    return $servers;
  }

  static public function resetMailServersCache ()
  {
    session::global_un_set('mailServers');
  }

  /*! \brief  Returns the available mailMethods
      @return Array   A list of all avaialable mailMethods_
   */
  static public function getMethodList (): array
  {
    global $class_mapping;
    $available = ['mailMethod' => 'mailMethod'];
    foreach ($class_mapping as $class => $path) {
      if (preg_match('/^mailMethod.+/', $class)) {
        $available[$class] = $class;
      }
    }
    return $available;
  }

  static public function getServerMethod (string $server): string
  {
    $methods = static::getMethodList();
    foreach ($methods as $class) {
      if ($class::serverMatch($server)) {
        @DEBUG(DEBUG_MAIL, __LINE__, __FUNCTION__, __FILE__, $class, 'Mail method for server "'.$server.'"');
        return $class;
      }
    }
    @DEBUG(DEBUG_MAIL, __LINE__, __FUNCTION__, __FILE__, '', 'No method found for server "'.$server.'"');
    return 'mailMethod';
  }

  static public function serverMatch (string $server): bool
  {
    return isset(static::getServerList()[$server]);
  }

  static public function getInstanceFromServer (string $server, $parent, string $type = NULL): mailMethod
  {
    /* Intialize the used mailMethod */
    if ($server == '') {
      $class = 'mailMethod';
    } else {
      $class = static::getServerMethod($server);
    }

    if ($parent instanceof mailAccount) {
      $type = 'user';
    } elseif ($type === NULL) {
      $type = 'group';
    }

    return new $class($parent, $type);
  }

  static public function getServerList (): array
  {
    global $config;
    $serverList = [];

    $ldap = $config->get_ldap_link();
    $ldap->cd($config->current['BASE']);
    $ldap->search('(objectClass=fdImapServer)', ['cn']);
    while ($attrs = $ldap->fetch()) {
      $serverList[$attrs['cn'][0]] = [
        'server_dn'     => $attrs['dn'],
        'connect'       => '',
        'admin'         => '',
        'password'      => '',
        'sieve_server'  => '',
        'sieve_option'  => '',
        'sieve_port'    => ''
      ];
    }

    return $serverList;
  }
}
