<?php
/*
  This code is part of FusionDirectory (http://www.fusiondirectory.org/)
  Copyright (C) 2007  Fabian Hicker
  Copyright (C) 2011-2021  FusionDirectory

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
*/

/*!
  \brief   mail plugin

  This class provides the functionality to read and write all attributes
  relevant for mail accounts from/to the LDAP.

  It does syntax checking and displays the formulars required.
  Special handling like sieve or imap actions will be implemented
  by the mailMethods.

 */
class mailAccount extends mailTabPlugin
{
  protected $displayHeader  = TRUE;

  static function plInfo (): array
  {
    return [
      'plShortName'     => _('Mail'),
      'plDescription'   => _('Mail settings'),
      'plFilter'        => '(mail=*)',
      'plIcon'          => 'geticon.php?context=applications&icon=internet-mail&size=48',
      'plSmallIcon'     => 'geticon.php?context=applications&icon=internet-mail&size=16',
      'plSelfModify'    => TRUE,
      'plPriority'      => 4,
      'plObjectClass'   => ['gosaMailAccount'],
      'plObjectType'    => ['user'],
      'plProvidedAcls'  => parent::generatePlProvidedAcls(static::getAttributesInfo()),
      'plForeignKeys'  => [
        'gosaMailServer' => [
          ['serverGeneric', 'cn'],
        ]
      ]
    ];
  }

  static function getAttributesInfo (): array
  {
    return  [
      'main' => [
        'name'  => _('Mail account'),
        'attrs' => [
          new MailAttribute(
            _('Primary address'), _('Primary mail address'),
            'mail', TRUE
          ),
          new SelectAttribute(
            _('Server'), _('Specify the mail server where the user will be hosted on'),
            'gosaMailServer', FALSE,
            []
          ),
          new MailQuotaAttribute(
            _('Quota size'), _('Define quota size in MiB'),
            'gosaMailQuota', FALSE,
            0, FALSE
          )
        ]
      ],
      'other_addresses' => [
        'name'  => _('Other addresses and redirections'),
        'attrs' => [
          new SetAttribute(
            new MailAttribute(
              _('Alternative addresses'), _('List of alternative mail addresses'),
              'gosaMailAlternateAddress'
            )
          ),
          new MailsAttribute(
            _('Forward messages to'), _('Addresses to which messages should be forwarded'),
            'gosaMailForwardingAddress'
          )
        ]
      ],
      'options1' => [
        'name' => _('Vacation message'),
        'attrs' => [
          new BooleanAttribute(
            _('Activate vacation message'),
            _('Select to automatically response with the vacation message defined below'),
            'flag_vacation', FALSE, FALSE, '',
            'V', ''
          ),
          new DateAttribute(
            _('from'), 'Starting date for vacation message',
            'gosaVacationStart', FALSE,
            'U'
          ),
          new DateAttribute(
            _('till'), 'Ending date for vacation message',
            'gosaVacationStop', FALSE,
            'U'
          ),
          new TextAreaAttribute(
            _('Vacation message'), 'The message you wish be sent during your absence',
            'gosaVacationMessage'
          ),
        ]
      ],
      'advanced' => [
        'name' => _('Advanced mail options'),
        'attrs' => [
          new FlagsAttribute('gosaMailDeliveryMode', [
            'flag_ownmailbox',
            'flag_vacation',
            'flag_localonly'
          ]),
          new BooleanAttribute(
            _('User is only allowed to send and receive local mails'),
            _('Select if user can only send and receive inside his own domain'),
            'flag_localonly', FALSE, FALSE, '',
            'L', ''
          ),
          new BooleanAttribute(
            _('No delivery to own mailbox'),
            _('Select if you want to forward mails without getting own copies of them'),
            'flag_ownmailbox', FALSE, FALSE, '',
            'I', ''
          ),
        ]
      ]
    ];
  }

  protected function loadAttributes ()
  {
    global $config;

    $this->attributesAccess['flag_ownmailbox']->setInLdap(FALSE);
    $this->attributesAccess['flag_vacation']->setInLdap(FALSE);
    $this->attributesAccess['flag_localonly']->setInLdap(FALSE);

    /* If mail can be used for login it must be unique */
    if (in_array('mail', explode(',', $config->get_cfg_value('loginAttribute')))) {
      $this->attributesAccess['mail']->setUnique('whole');
    }

    parent::loadAttributes();
  }

  /* We need $mailAccount->uid to return the uid for mailMethod */
  public function __get ($name)
  {
    if (in_array($name, ['cn','uid']) && isset($this->parent)) {
      return $this->parent->getBaseObject()->$name;
    } elseif (in_array($name, ['uidNumber','gidNumber']) && isset($this->parent)) {
      return $this->parent->by_object['posixAccount']->$name;
    } else {
      /* Calling default behaviour */
      return parent::__get($name);
    }
  }

  public function __isset ($name)
  {
    if (in_array($name, ['cn','uid']) && isset($this->parent)) {
      return isset($this->parent->getBaseObject()->$name);
    } elseif (in_array($name, ['uidNumber','gidNumber']) && isset($this->parent)) {
      return isset($this->parent->by_object['posixAccount']->$name);
    } else {
      /* Calling default behaviour */
      return parent::__isset($name);
    }
  }

  public function mailServerChanged ()
  {
    parent::mailServerChanged();

    /* Initialize configured values */
    if (!$this->is_template && $this->initially_was_account) {
      try {
        if ($this->mailMethod->accountExists()) {
          /* Read quota */
          $this->attributesAccess['gosaMailQuota']->setMailMethod($this->mailMethod);
        }
      } catch (MailMethodError $e) {
        $error = new SimplePluginError(
          $this,
          htmlescape(sprintf(
            _('Mail method error: %s'),
            $e->getMessage()
          )),
          0,
          $e
        );
        $error->display();
      }

      $this->attributesAccess['mail']->setDisabled(!$this->mailMethod->isModifyableMail());
      $this->attributesAccess['gosaMailServer']->setDisabled(!$this->mailMethod->isModifyableServer());
    }

    /* Feature switches */

    if ($this->mailMethod->localOnlyEnabled()) {
      $this->attributesAccess['flag_localonly']->setDisabled(FALSE);
      $this->attributesAccess['flag_localonly']->setVisible(TRUE);
    } else {
      $this->attributesAccess['flag_localonly']->setDisabled(TRUE);
      $this->attributesAccess['flag_localonly']->setVisible(FALSE);
      $this->attributesAccess['flag_localonly']->setValue(FALSE);
    }

    if ($this->mailMethod->forwardOnlyEnabled()) {
      $this->attributesAccess['flag_ownmailbox']->setDisabled(FALSE);
    } else {
      $this->attributesAccess['flag_ownmailbox']->setDisabled(TRUE);
      $this->attributesAccess['flag_ownmailbox']->setValue(FALSE);
    }

    if ($this->mailMethod->vacationEnabled() && $this->mailMethod->vacationRangeEnabled()) {
      $this->attributesAccess['gosaVacationStop']->setDisabled(FALSE);
      $this->attributesAccess['gosaVacationStart']->setDisabled(FALSE);
      $this->attributesAccess['flag_vacation']->setManagedAttributes(
        [
          'erase' => [
            FALSE => [
              'gosaVacationStart','gosaVacationStop','gosaVacationMessage'
            ]
          ]
        ]
      );
    } else {
      $this->attributesAccess['gosaVacationStart']->setDisabled(TRUE);
      $this->attributesAccess['gosaVacationStart']->setVisible(FALSE);
      $this->attributesAccess['gosaVacationStart']->setValue('');
      $this->attributesAccess['gosaVacationStop']->setDisabled(TRUE);
      $this->attributesAccess['gosaVacationStop']->setVisible(FALSE);
      $this->attributesAccess['gosaVacationStop']->setValue('');
      if ($this->mailMethod->vacationEnabled()) {
        $this->attributesAccess['flag_vacation']->setDisabled(FALSE);
        $this->attributesAccess['flag_vacation']->setVisible(TRUE);
        $this->attributesAccess['gosaVacationMessage']->setDisabled(FALSE);
        $this->attributesAccess['gosaVacationMessage']->setVisible(TRUE);
        $this->attributesAccess['flag_vacation']->setManagedAttributes(
          [
            'erase' => [
              FALSE => [
                'gosaVacationMessage'
              ]
            ]
          ]
        );
      } else {
        $this->attributesAccess['flag_vacation']->setDisabled(TRUE);
        $this->attributesAccess['flag_vacation']->setVisible(FALSE);
        $this->attributesAccess['flag_vacation']->setValue(FALSE);
        $this->attributesAccess['flag_vacation']->setManagedAttributes([]);
        $this->attributesAccess['gosaVacationMessage']->setDisabled(TRUE);
        $this->attributesAccess['gosaVacationMessage']->setVisible(FALSE);
        $this->attributesAccess['gosaVacationMessage']->setValue('');
      }
    }
  }

  protected function mailAccountPostUpdate ()
  {
    try {
      $this->mailMethod->setQuota($this->gosaMailQuota);
    } catch (MailMethodError $e) {
      $error = new SimplePluginError(
        $this,
        htmlescape(sprintf(
          _('Cannot write quota settings: %s'),
          $e->getMessage()
        )),
        0,
        $e
      );
      $error->display();
    }

    if (!is_integer(strpos($this->gosaMailDeliveryMode, 'C'))) {
      /* Do not write sieve settings if this account is new and
         doesn't seem to exist. */
      try {
        if (!$this->initially_was_account && !$this->mailMethod->accountExists()) {
          @DEBUG(DEBUG_MAIL, __LINE__, __FUNCTION__, __FILE__,
              'Skipping sieve settings, the account doesn’t seem to be created already.', '');
        } else {
          $this->mailMethod->saveSieveSettings();
        }
      } catch (MailMethodError $e) {
        $error = new SimplePluginError(
          $this,
          htmlescape(sprintf(
            _('Mail error saving sieve settings: %s'),
            $e->getMessage()
          )),
          0,
          $e
        );
        $error->display();
      }
    } else {
      if ($this->sieve_management) {
        @DEBUG(DEBUG_MAIL, __LINE__, __FUNCTION__, __FILE__,
            'User uses an own sieve script, skipping sieve update.', '');
        $this->sieve_management->save();
      }
    }
  }
}
