<?php
/*
  This code is part of FusionDirectory (http://www.fusiondirectory.org/)
  Copyright (C) 2007  Fabian Hicker
  Copyright (C) 2011-2018  FusionDirectory

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
*/

/*!
 * \brief mail tab plugin
 *
 * This class provides the functionality to handle a mailMethod instance from a simplePlugin tab
 */
class mailTabPlugin extends simplePlugin
{
  protected $serverAttr = 'gosaMailServer';

  private $mainSectionAttrs = [];

  public $mailMethod        = NULL;
  public $initialMailMethod = NULL;

  function __construct (string $dn = NULL, $object = NULL, $parent = NULL, bool $mainTab = FALSE, array $attributesInfo = NULL)
  {
    global $config;
    parent::__construct($dn, $object, $parent, $mainTab, $attributesInfo);

    $this->attributesAccess[$this->serverAttr]->setChoices(array_keys(mailMethod::getMailServers()));
    $this->attributesAccess[$this->serverAttr]->setSubmitForm('mailServerChanged');
    $this->mailServerChanged();
  }

  protected function loadAttributes ()
  {
    $this->mainSectionAttrs = array_keys($this->attributesInfo['main']['attrs']);

    parent::loadAttributes();
  }

  function foreignKeyUpdate (string $field, $oldvalue, $newvalue, array $source)
  {
    if ($field == $this->serverAttr) {
      mailMethod::resetMailServersCache();
      $this->attributesAccess[$this->serverAttr]->setChoices(array_keys(mailMethod::getMailServers()));
    }
    parent::foreignKeyUpdate($field, $oldvalue, $newvalue, $source);
  }

  /*!
   * \brief This function is called on the copied object to set its dn to where it will be saved
   */
  function resetCopyInfos ()
  {
    parent::resetCopyInfos();
    $this->initialMailMethod = NULL;
  }

  public function mailServerChanged ()
  {
    $this->mailMethod = mailMethod::getInstanceFromServer($this->attributesAccess[$this->serverAttr]->getValue(), $this);
    if ($this->initialMailMethod === NULL) {
      $this->initialMailMethod = $this->mailMethod;
    }

    /* Add attributes from mail method */
    foreach (array_keys($this->attributesInfo['main']['attrs']) as $oldattr) {
      if (!in_array($oldattr, $this->mainSectionAttrs)) {
        $this->removeAttribute('main', $oldattr);
      }
    }
    $addAttributes = $this->mailMethod->additionalAttributes($this->attributesAccess[$this->serverAttr]->getAcl());
    foreach ($addAttributes as $attribute) {
      $attribute->setInLdap(FALSE);
      $this->addAttribute(
        'main',
        $attribute
      );
    }
  }

  function check (): array
  {
    $errors = parent::check();

    if (!$this->is_template && is_object($this->mailMethod)) {
      $errors = array_merge($errors, $this->mailMethod->check());
    }

    return $errors;
  }

  protected function shouldSave (): bool
  {
    /* mail method might have something to save (like password change from main tab for instance) */
    return TRUE;
  }

  /*!
  *  \brief Should the mail method create the mailbox? (mail server change or tab activation)
  */
  public function isMailboxCreation (): bool
  {
    return (!$this->initially_was_account || $this->attributesAccess[$this->serverAttr]->hasChanged());
  }

  protected function mailAccountPostUpdate ()
  {
  }

  public function ldap_save (): array
  {
    if (!empty($this->attrs)) {
      $errors = parent::ldap_save();
      if (!empty($errors)) {
        return $errors;
      }
    }

    $errors = [];

    /* Only do mail actions if we are not a template */
    if (!$this->is_template) {
      if ($this->initially_was_account && $this->attributesAccess[$this->serverAttr]->hasChanged()) {
        if (
          ($this->initialMailMethod === NULL) ||
          (get_class($this->initialMailMethod) !== mailMethod::getServerMethod($this->attributesAccess[$this->serverAttr]->getInitialValue()))
        ) {
          $errors[] = new SimplePluginError(
            $this,
            htmlescape(
              sprintf(
                'Initial method wrong or missing: %s !== %s',
                ($this->initialMailMethod === NULL ? 'NULL' : get_class($this->initialMailMethod)),
                mailMethod::getServerMethod($this->attributesAccess[$this->serverAttr]->getInitialValue())
              )
            )
          );
        } else {
          try {
            $this->initialMailMethod->deleteMailbox();
          } catch (MailMethodError $e) {
            $errors[] = new SimplePluginError(
              $this,
              htmlescape(sprintf(_('Cannot remove mailbox: %s'), $e->getMessage())),
              0,
              $e
            );
          }
        }
      }
      try {
        $this->mailMethod->updateMailbox();
        $this->mailAccountPostUpdate();
      } catch (MailMethodError $e) {
        $errors[] = new SimplePluginError(
          $this,
          htmlescape(sprintf(_('Cannot update mailbox: %s'), $e->getMessage())),
          0,
          $e
        );
        if (!$this->initially_was_account) {
          /* Mail method failed to create, cancel ldap save */
          $this->prepareNextCleanup();
          $this->prepare_remove();
          $this->ldap_remove();
        }
      }
      $this->mailMethod->disconnect();
    }

    return $errors;
  }

  function post_remove ()
  {
    /* Let the mailMethod remove this mailbox */
    if (!$this->is_template) {
      try {
        $this->initialMailMethod->deleteMailbox();
      } catch (MailMethodError $e) {
        $error = new SimplePluginError(
          $this,
          htmlescape(sprintf(_('Cannot remove mailbox: %s'), $e->getMessage())),
          0,
          $e
        );
        $error->display();
      }
      $this->initialMailMethod->disconnect();
    }
    parent::post_remove();
  }
}
